import pytest

from lion_core.communication.assistant_response import AssistantResponse
from lion_core.communication.message import MessageFlag, MessageRole
from lion_core.sys_utils import SysUtil


# Basic functionality tests
def test_assistant_response_init():
    response = AssistantResponse({"content": "Test response"}, "assistant", "user")
    assert response.role == MessageRole.ASSISTANT
    assert response.sender == "assistant"
    assert response.recipient == "user"
    assert response.response == "Test response"


def test_assistant_response_init_default_sender():
    response = AssistantResponse({"content": "Test response"}, None, "user")
    assert response.sender == "N/A"


def test_assistant_response_init_with_message_load():
    protected_params = {
        "role": MessageRole.ASSISTANT,
        "content": {"assistant_response": "Test"},
        "sender": "assistant",
    }
    response = AssistantResponse(
        MessageFlag.MESSAGE_LOAD,
        MessageFlag.MESSAGE_LOAD,
        MessageFlag.MESSAGE_LOAD,
        protected_init_params=protected_params,
    )
    assert response.role == MessageRole.ASSISTANT
    assert response.response == "Test"


def test_assistant_response_init_with_message_clone():
    response = AssistantResponse(
        MessageFlag.MESSAGE_CLONE, MessageFlag.MESSAGE_CLONE, MessageFlag.MESSAGE_CLONE
    )
    assert response.role == MessageRole.ASSISTANT


# Edge cases and additional tests
def test_assistant_response_empty_content():
    response = AssistantResponse({"content": ""}, "assistant", "user")
    assert response.response == ""


def test_assistant_response_missing_content():
    response = AssistantResponse({}, "assistant", "user")
    assert response.response == ""


def test_assistant_response_none_content():
    response = AssistantResponse(None, "assistant", "user")
    assert response.response is ""


def test_assistant_response_very_long_content():
    long_content = "a" * 10000
    response = AssistantResponse({"content": long_content}, "assistant", "user")
    assert len(response.response) == 10000


def test_assistant_response_with_special_characters():
    special_chars = "!@#$%^&*()_+{}[]|\\:;\"'<>,.?/~`"
    response = AssistantResponse({"content": special_chars}, "assistant", "user")
    assert response.response == special_chars


def test_assistant_response_with_unicode():
    unicode_content = "你好世界 - Hello World - こんにちは世界"
    response = AssistantResponse({"content": unicode_content}, "assistant", "user")
    assert response.response == unicode_content


def test_assistant_response_serialization():
    response = AssistantResponse({"content": "Test response"}, "assistant", "user")
    serialized = response.to_dict()
    deserialized = AssistantResponse.from_dict(serialized)
    assert deserialized.response == response.response
    assert deserialized.sender == response.sender


def test_assistant_response_clone():
    original = AssistantResponse({"content": "Original response"}, "assistant", "user")
    cloned = original.clone()
    assert cloned.response == original.response
    assert cloned.ln_id != original.ln_id


# Performance test
def test_assistant_response_performance():
    import time

    start_time = time.time()
    for _ in range(1000):
        AssistantResponse({"content": "Test response"}, "assistant", "user")
    end_time = time.time()
    assert (
        end_time - start_time < 1
    )  # Should create 1000 AssistantResponse objects in less than 1 second


# Thread safety test
def test_assistant_response_thread_safety():
    import threading

    def create_response():
        AssistantResponse(
            {"content": f"Thread {threading.get_ident()}"}, "assistant", "user"
        )

    threads = [threading.Thread(target=create_response) for _ in range(100)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()


# Test with invalid UTF-8 sequences
def test_assistant_response_invalid_utf8():
    invalid_utf8 = b"Invalid UTF-8: \xff"
    with pytest.raises(UnicodeDecodeError):
        AssistantResponse(
            {"content": invalid_utf8.decode("utf-8")}, "assistant", "user"
        )


# Test assistant response with various content types
@pytest.mark.parametrize(
    "content",
    [
        123,
        3.14,
        True,
        ["list", "item"],
        {"nested": "dict"},
    ],
)
def test_assistant_response_various_content_types(content):
    response = AssistantResponse({"content": content}, "assistant", "user")
    assert response.response == content


# Test for potential memory leaks
def test_assistant_response_memory_usage():
    import sys

    initial_size = sys.getsizeof(
        AssistantResponse({"content": "Test"}, "assistant", "user")
    )
    large_response = AssistantResponse({"content": "a" * 1000000}, "assistant", "user")
    large_size = sys.getsizeof(large_response)
    assert (
        large_size - initial_size < 2000000
    )  # Ensure no significant unexpected memory increase


# Test assistant response with empty strings for all parameters
def test_assistant_response_empty_strings():
    response = AssistantResponse({"content": ""}, None, None)
    assert response.response == ""
    assert response.sender == "N/A"
    assert response.recipient == "N/A"


# Test assistant response with None values for sender and recipient
def test_assistant_response_none_values():
    response = AssistantResponse({"content": "Test"}, None, None)
    assert response.sender == "N/A"
    assert response.recipient == "N/A"


# Test assistant response with various types in the content dictionary
def test_assistant_response_complex_content():
    complex_content = {
        "content": "Main content",
        "additional_info": {
            "nested": ["list", "of", "items"],
            "number": 42,
            "boolean": True,
        },
    }
    response = AssistantResponse(complex_content, "assistant", "user")
    assert response.response == "Main content"


# Test assistant response with missing 'content' key
def test_assistant_response_missing_content_key():
    response = AssistantResponse({"other_key": "value"}, "assistant", "user")
    assert response.response == {"other_key": "value"}


# Test assistant response with non-dict input
def test_assistant_response_non_dict_input():
    response = AssistantResponse("Just a string", "assistant", "user")
    assert response.response == "Just a string"


# Test assistant response with very large input
def test_assistant_response_very_large_input():
    large_input = {"content": "a" * 1000000}  # 1MB of content
    response = AssistantResponse(large_input, "assistant", "user")
    assert len(response.response) == 1000000



from typing import Any

from typing_extensions import override

from lion_core.communication.message import MessageFlag, MessageRole, RoledMessage
from lion_core.generic.note import Note
from lion_core.sys_utils import SysUtil

DEFAULT_SYSTEM = "You are a helpful AI assistant. Let's think step by step."


def format_system_content(
    system_datetime: bool | str | None,
    system_message: str,
) -> Note:
    """
    Format the system content with optional datetime information.

    This function creates a system message formatted as a `Note` object, optionally
    including a timestamp. The timestamp can be automatically generated in ISO format
    or provided as a string.

    Args:
        system_datetime (bool | str | None): A flag or string indicating whether to include
            a datetime in the system message.
            - If `False` or `None`, no datetime is included.
            - If `True`, the current system datetime is appended in ISO format.
            - If a string is provided, it is appended as the datetime.
        system_message (str): The content of the system message. If not provided,
            a default system message is used.

    Returns:
        Note: A `Note` object containing the formatted system content with the optional datetime.

    Example:
        >>> format_system_content(True, "System maintenance complete")
        Note with content: "System maintenance complete. System Date: 2024-08-19T12:34"

        >>> format_system_content("2024-08-19", "System update scheduled")
        Note with content: "System update scheduled. System Date: 2024-08-19"
    """
    system_message = system_message or DEFAULT_SYSTEM
    if not system_datetime:
        return Note(system_info=str(system_message))
    if isinstance(system_datetime, str):
        return Note(system_info=f"{system_message}. System Date: {system_datetime}")
    if system_datetime:
        date = SysUtil.time(type_="iso", timespec="minutes")
        return Note(system_info=f"{system_message}. System Date: {date}")


class System(RoledMessage):
    """
    Represents a system message in a language model (LLM) conversation.

    The `System` class is a specialized type of `RoledMessage` that encapsulates
    messages from the system within a conversation. These messages may include
    system-related information such as updates, notifications, or status reports.

    Inherits:
        - RoledMessage: Provides basic message structure, including content and role.

    Attributes:
        content (Note): The content of the system message, formatted with optional datetime.
        role (MessageRole): The role of the message, set to "system".

    Properties:
        system_info (Any): Retrieves the system information stored in the message content.

    Methods:
        __init__(self, system, sender, recipient, system_datetime, protected_init_params):
            Initializes a new `System` message instance.
    """

    @override
    def __init__(
        self,
        system: Any | MessageFlag = None,
        sender: str | None | MessageFlag = None,
        recipient: str | None | MessageFlag = None,
        system_datetime: bool | str | None | MessageFlag = None,
        protected_init_params: dict | None = None,
    ):
        """
        Initialize a `System` message instance.

        This constructor initializes a `System` message with optional content, sender,
        recipient, and datetime information. The system message can be created fresh,
        loaded from a previous state, or cloned from an existing instance.

        Args:
            system (Any | MessageFlag, optional): The main content of the system message.
                Defaults to None.
            sender (str | None | MessageFlag, optional): The sender of the system message.
                Defaults to None.
            recipient (str | None | MessageFlag, optional): The intended recipient of the system message.
                Defaults to None.
            system_datetime (bool | str | None | MessageFlag, optional): A flag or string to include
                datetime in the message. Defaults to None.
            protected_init_params (dict | None, optional): Protected initialization parameters
                for message loading. Defaults to None.
        """
        if all(
            x == MessageFlag.MESSAGE_LOAD
            for x in [system, sender, recipient, system_datetime]
        ):
            super().__init__(**protected_init_params)
            return

        if all(
            x == MessageFlag.MESSAGE_CLONE
            for x in [system, sender, recipient, system_datetime]
        ):
            super().__init__(role=MessageRole.SYSTEM)
            return

        super().__init__(
            role=MessageRole.SYSTEM,
            sender=sender or "system",
            content=format_system_content(
                system_datetime=system_datetime, system_message=system
            ),
            recipient=recipient or "N/A",
        )

    @property
    def system_info(self) -> Any:
        """
        Retrieve the system information stored in the message content.

        This property accesses the `system_info` field in the message content,
        which typically contains the formatted system message along with any
        datetime information.

        Returns:
            Any: The system information stored in the message content.
        """
        return self.content.get("system_info", None)


# File: lion_core/communication/system.py



import pytest

from lion_core.libs.data_handlers._nget import nget
from lion_core.setting import LN_UNDEFINED


@pytest.mark.parametrize(
    "data, indices, expected",
    [
        ({"a": {"b": {"c": 3}}}, ["a", "b", "c"], 3),
        ([1, [2, [3, 4]]], [1, 1, 0], 3),
        ({"a": [1, {"b": 2}]}, ["a", 1, "b"], 2),
        ({"a": 1, "b": 2}, ["b"], 2),
        ({"a": {"b": [{"c": 1}, {"c": 2}]}}, ["a", "b", 1, "c"], 2),
        ({"a": {"b": {"c": {"d": {"e": 5}}}}}, ["a", "b", "c", "d", "e"], 5),
        ([[[1]]], [0, 0, 0], 1),
        ({"a": [1, 2, {"b": [3, 4, {"c": 5}]}]}, ["a", 2, "b", 2, "c"], 5),
    ],
)
def test_nget_valid_paths(data, indices, expected):
    assert nget(data, indices) == expected


@pytest.mark.parametrize(
    "data, indices, default, expected",
    [
        ({"a": {"b": 2}}, ["a", "c"], 10, 10),
        ({"a": [1, 2, 3]}, [0, 1], "default", "default"),
        ({}, ["a", "b"], None, None),
        ([], [0], LN_UNDEFINED, LN_UNDEFINED),
    ],
)
def test_nget_with_default(data, indices, default, expected):
    assert nget(data, indices, default) == expected


@pytest.mark.parametrize(
    "data, indices",
    [
        ({}, ["a"]),
        ([], [0]),
        ({"a": {"b": 2}}, ["a", "c"]),
        ([1, 2, 3], [3]),
        ({"a": [1, 2]}, ["a", 2]),
        ({"a": {"b": 1}}, ["a", "b", "c"]),
    ],
)
def test_nget_raises_lookup_error(data, indices):
    with pytest.raises(LookupError):
        nget(data, indices)


@pytest.mark.parametrize(
    "data, indices",
    [
        ([1, 2, 3], ["a"]),
        ({"a": 1}, [0]),
        ({"a": [1, 2, 3]}, ["a", "b"]),
    ],
)
def test_nget_raises_type_error(data, indices):
    with pytest.raises(LookupError):
        nget(data, indices)


def test_nget_empty_indices():
    with pytest.raises(ValueError, match="Indices list cannot be empty"):
        nget({"a": 1}, [])


def test_nget_none_data():
    with pytest.raises(TypeError):
        nget(None, ["a"])


def test_nget_non_subscriptable():
    with pytest.raises(LookupError):
        nget(42, [0])


def test_nget_with_zero_index():
    data = [1, 2, 3]
    assert nget(data, [0]) == 1


def test_nget_with_negative_index():
    data = [1, 2, 3]
    with pytest.raises(LookupError):
        nget(data, [-1])


def test_nget_with_string_index_for_list():
    data = [1, 2, 3]
    with pytest.raises(LookupError):
        nget(data, ["0"])


def test_nget_with_int_index_for_dict():
    data = {"0": "value"}
    assert nget(data, ["0"]) == "value"


def test_nget_with_nested_default_dict():
    from collections import defaultdict

    data = defaultdict(lambda: defaultdict(int))
    data["a"]["b"] = 1
    assert nget(data, ["a", "b"]) == 1
    assert nget(data, ["a", "c"]) == 0


def test_nget_with_custom_object():
    class CustomObj:
        def __init__(self):
            self.attr = {"key": "value"}

        def __getitem__(self, key):
            return self.attr[key]

    obj = CustomObj()
    assert nget(obj, ["key"]) == "value"


def test_nget_with_property():
    class PropObj:
        @property
        def prop(self):
            return {"key": "value"}

    obj = PropObj()
    assert nget(obj, ["prop", "key"]) == "value"


@pytest.mark.parametrize(
    "data, indices, expected",
    [
        ({"a": 1, "b": 2, "c": 3}, ["a", "b", "c"], 3),
        ([1, 2, [3, 4, [5, 6]]], [2, 2, 1], 6),
        ({"a": [{"b": {"c": [1, 2, 3]}}]}, ["a", 0, "b", "c", 2], 3),
    ],
)
def test_nget_long_paths(data, indices, expected):
    assert nget(data, indices) == expected


def test_nget_with_large_nested_structure():
    large_data = {"level1": {}}
    current = large_data["level1"]
    for i in range(2, 101):
        current[f"level{i}"] = {}
        current = current[f"level{i}"]
    current["value"] = "deep"

    indices = [f"level{i}" for i in range(1, 101)] + ["value"]
    assert nget(large_data, indices) == "deep"


def test_nget_performance_with_large_list(benchmark):
    large_list = list(range(10**6))

    def fetch_last():
        return nget(large_list, [-1])

    result = benchmark(fetch_last)
    assert result == 999999


def test_nget_with_all_python_basic_types():
    data = {
        "int": 1,
        "float": 2.0,
        "complex": 1 + 2j,
        "str": "string",
        "list": [1, 2, 3],
        "tuple": (4, 5, 6),
        "dict": {"key": "value"},
        "set": {7, 8, 9},
        "frozenset": frozenset([10, 11, 12]),
        "bool": True,
        "none": None,
        "bytes": b"bytes",
        "bytearray": bytearray(b"bytearray"),
    }
    for key in data.keys():
        assert nget(data, [key]) == data[key]


@pytest.mark.parametrize(
    "data, indices, expected",
    [
        ({"a": {"b": {"c": 3}}}, "a.b.c", 3),
        ([1, [2, [3, 4]]], "1.1.0", 3),
        ({"a": [1, {"b": 2}]}, "a.1.b", 2),
    ],
)
def test_nget_with_string_indices(data, indices, expected):
    assert nget(data, indices.split(".")) == expected


# File: tests/test_nget.py



import asyncio
import unittest
from unittest.mock import AsyncMock, patch

from lion_core.libs.function_handlers._mcall import mcall


async def async_func(x: int) -> int:
    await asyncio.sleep(0.1)
    return x * 2


def sync_func(x: int) -> int:
    return x * 2


async def async_func_with_error(x: int) -> int:
    await asyncio.sleep(0.1)
    if x == 3:
        raise ValueError("mock error")
    return x * 2


def sync_func_with_error(x: int) -> int:
    if x == 3:
        raise ValueError("mock error")
    return x * 2


async def mock_handler(e: Exception) -> str:
    return f"handled: {str(e)}"


class TestMCallFunction(unittest.IsolatedAsyncioTestCase):
    async def test_mcall_single_func(self):
        inputs = [1, 2, 3, 4, 5]
        result = await mcall(inputs, async_func)
        self.assertEqual(result, [2, 4, 6, 8, 10])

    async def test_mcall_multiple_funcs_explode(self):
        inputs = [1, 2, 3, 4, 5]
        funcs = [async_func, async_func]
        results = await mcall(inputs, funcs, explode=True)
        self.assertEqual(results, [[2, 4, 6, 8, 10], [2, 4, 6, 8, 10]])

    async def test_mcall_multiple_funcs_no_explode(self):
        inputs = [1, 2, 3, 4, 5]
        funcs = [async_func, async_func, async_func, async_func, async_func]
        result = await mcall(inputs, funcs)
        self.assertEqual(result, [2, 4, 6, 8, 10])

    async def test_mcall_with_retries(self):
        inputs = [1, 2, 3, 4, 5]
        funcs = [async_func_with_error] * 5
        results = await mcall(inputs, funcs, retries=1, default=0)
        self.assertEqual(results, [2, 4, 0, 8, 10])

    async def test_mcall_with_timeout(self):
        inputs = [1, 2, 3]
        with self.assertRaises(RuntimeError):
            await mcall(inputs, async_func, timeout=0.05)

    async def test_mcall_with_error_handling(self):
        error_map = {ValueError: mock_handler}
        inputs = [1, 2, 3, 4, 5]
        funcs = [async_func_with_error] * 5
        result = await mcall(inputs, funcs, error_map=error_map)
        self.assertEqual(result, [2, 4, "handled: mock error", 8, 10])

    async def test_mcall_with_max_concurrent(self):
        inputs = [1, 2, 3, 4, 5]
        result = await mcall(inputs, async_func, max_concurrent=1)
        self.assertEqual(result, [2, 4, 6, 8, 10])

    async def test_mcall_with_throttle_period(self):
        inputs = [1, 2, 3, 4, 5]
        result = await mcall(inputs, async_func, throttle_period=0.2)
        self.assertEqual(result, [2, 4, 6, 8, 10])

    async def test_mcall_with_initial_delay(self):
        inputs = [1, 2, 3]
        with patch("asyncio.sleep", new_callable=AsyncMock) as mock_sleep:
            result = await mcall(inputs, async_func, initial_delay=0.5)
            mock_sleep.assert_any_call(0.5)
            self.assertEqual(result, [2, 4, 6])

    async def test_mcall_with_kwargs(self):
        async def async_func_with_kwargs(x: int, add: int) -> int:
            await asyncio.sleep(0.1)
            return x + add

        inputs = [1, 2, 3, 4, 5]
        result = await mcall(inputs, async_func_with_kwargs, add=10)
        self.assertEqual(result, [11, 12, 13, 14, 15])


if __name__ == "__main__":
    unittest.main()



from typing import Any, Callable, ClassVar, Literal

from pydantic import Field, model_validator
from typing_extensions import override

from lion_core.abc import BaseiModel
from lion_core.action import Tool, ToolManager
from lion_core.communication.action_request import ActionRequest
from lion_core.communication.action_response import ActionResponse
from lion_core.communication.assistant_response import AssistantResponse
from lion_core.communication.instruction import Instruction
from lion_core.communication.mail import Mail
from lion_core.communication.message import MessageFlag, RoledMessage
from lion_core.communication.package import Package
from lion_core.communication.system import System
from lion_core.converter import ConverterRegistry
from lion_core.generic.exchange import Exchange
from lion_core.generic.note import Note
from lion_core.generic.pile import Pile
from lion_core.generic.progression import Progression, prog
from lion_core.libs import is_same_dtype
from lion_core.session.base import BaseSession
from lion_core.session.msg_handlers.create_msg import create_message
from lion_core.session.msg_handlers.validate_msg import validate_message


class BranchConverterRegistry(ConverterRegistry):
    """Registry for Branch converters."""

    pass


class Branch(BaseSession):
    """
    Represents a branch in the conversation tree with tools and messages.

    This class manages a conversation branch, including messages, tools,
    and communication within the branch.
    """

    messages: Pile | None = Field(None)
    tool_manager: ToolManager | None = Field(None)
    mailbox: Exchange | None = Field(None)
    progress: Progression | None = Field(None)

    _converter_registry: ClassVar = BranchConverterRegistry

    @model_validator(mode="before")
    def _validate_input(cls, data: dict) -> dict:
        messages = data.pop("messages", None)
        data["messages"] = cls.pile_type(
            validate_message(messages),
            {RoledMessage},
            strict=False,
        )
        data["progress"] = prog(
            data.pop(list(data["messages"]), []),
        )
        data["tool_manager"] = data.pop(
            data["tool_manager"],
            ToolManager(),
        )
        data["mailbox"] = data.pop(
            data["mailbox"],
            Exchange(),
        )
        if "tools" in data:
            data["tool_manager"].register_tools(data.pop("tools"))
        return cls.validate_system(data)

    @model_validator(mode="after")
    def _check_system(self):
        if self.system not in self.messages:
            self.messages.include(self.system)
            self.progress.insert(0, self.system)
        return self

    def set_system(self, system: System) -> None:
        """
        Set or update the system message for the branch.

        Args:
            system: The new system message.
        """
        if len(self.progress) < 1:
            self.messages.include(system)
            self.system = system
            self.progress[0] = self.system
        else:
            self._change_system(system, delete_previous_system=True)
            self.progress[0] = self.system

    def add_message(
        self,
        *,
        sender: Any | MessageFlag = None,
        recipient: Any | MessageFlag = None,
        instruction: Any | MessageFlag = None,
        context: Any | MessageFlag = None,
        guidance: Any | MessageFlag = None,
        request_fields: dict | MessageFlag = None,
        system: Any = None,
        system_sender: Any = None,
        system_datetime: bool | str | None | MessageFlag = None,
        images: list | MessageFlag = None,
        image_detail: Literal["low", "high", "auto"] | MessageFlag = None,
        assistant_response: str | dict | None | MessageFlag = None,
        action_request: ActionRequest | None = None,
        action_response: ActionResponse | None = None,
        func: str | Callable | MessageFlag = None,
        arguments: dict | MessageFlag = None,
        func_output: Any | MessageFlag = None,
        metadata: Note | dict = None,  # additional branch parameters
        delete_previous_system: bool = None,
    ) -> bool:
        _msg = create_message(
            sender=sender,
            recipient=recipient,
            instruction=instruction,
            context=context,
            guidance=guidance,
            request_fields=request_fields,
            system=system,
            system_datetime=system_datetime,
            images=images,
            image_detail=image_detail,
            assistant_response=assistant_response,
            action_request=action_request,
            action_response=action_response,
            func=func,
            arguments=arguments,
            func_output=func_output,
            system_sender=system_sender,
        )

        if isinstance(_msg, System):
            _msg.recipient = self.ln_id  # the branch itself, system is to the branch
            self._change_system(
                system=_msg,
                delete_previous_system=delete_previous_system,
            )

        if isinstance(_msg, Instruction):
            _msg.sender = sender or self.user
            _msg.recipient = recipient or self.ln_id

        if isinstance(_msg, AssistantResponse):
            _msg.sender = sender or self.ln_id
            _msg.recipient = recipient or self.user or "user"

        if isinstance(_msg, ActionRequest):
            _msg.sender = sender or self.ln_id
            _msg.recipient = recipient or "N/A"

        if isinstance(_msg, ActionResponse):
            _msg.sender = sender or "N/A"
            _msg.recipient = recipient or self.ln_id

        if metadata:
            _msg.metadata.update(metadata, ["extra"])

        return self.messages.include(_msg)

    def clear_messages(self) -> None:
        """Clear all messages except the system message."""
        self.messages.clear()
        self.progress.clear()
        self.messages.include(self.system)
        self.progress.insert(0, self.system)

    def _change_system(
        self,
        system: System,
        delete_previous_system: bool = False,
    ):
        """
        Change the system message.

        Args:
            system: The new system message.
            delete_previous_system: If True, delete the previous system message.
        """
        old_system = self.system
        self.system = system
        self.messages.insert(0, self.system)

        if delete_previous_system:
            self.messages.exclude(old_system)

    def send(
        self,
        recipient: str,
        category: str,
        package: Any,
        request_source: str,
    ) -> None:
        """
        Send a mail to a recipient.

        Args:
            recipient: The recipient's ID.
            category: The category of the mail.
            package: The content of the mail.
            request_source: The source of the request.
        """
        package = Package(
            category=category,
            package=package,
            request_source=request_source,
        )

        mail = Mail(
            sender=self.ln_id,
            recipient=recipient,
            package=package,
        )
        self.mailbox.include(mail, "out")

    def receive(
        self,
        sender: str,
        message: bool = False,
        tool: bool = False,
        imodel: bool = False,
    ) -> None:
        """
        Receives mail from a sender.

        Args:
            sender (str): The ID of the sender.
            message (bool, optional): Whether to process message mails. Defaults to True.
            tool (bool, optional): Whether to process tool mails. Defaults to True.
            imodel (bool, optional): Whether to process imodel mails. Defaults to True.

        Raises:
            ValueError: If the sender does not exist or the mail category is invalid.
        """
        skipped_requests = prog()
        if sender not in self.mailbox.pending_ins.keys():
            raise ValueError(f"No package from {sender}")
        while self.mailbox.pending_ins[sender].size() > 0:
            mail_id = self.mailbox.pending_ins[sender].popleft()
            mail: Mail = self.mailbox.pile[mail_id]

            if mail.category == "message" and message:
                if not isinstance(mail.package.package, RoledMessage):
                    raise ValueError("Invalid message format")
                new_message = mail.package.package.clone()
                new_message.sender = mail.sender
                new_message.recipient = self.ln_id
                self.messages.include(new_message)
                self.mailbox.pile.pop(mail_id)

            elif mail.category == "tool" and tool:
                if not isinstance(mail.package.package, Tool):
                    raise ValueError("Invalid tools format")
                self.tool_manager.register_tools(mail.package.package)
                self.mailbox.pile.pop(mail_id)

            elif mail.category == "imodel" and imodel:
                if not isinstance(mail.package.package, BaseiModel):
                    raise ValueError("Invalid iModel format")
                self.imodel = mail.package.package
                self.mailbox.pile.pop(mail_id)

            else:
                skipped_requests.append(mail)

        self.mailbox.pending_ins[sender] = skipped_requests

        if self.mailbox.pending_ins[sender].size() == 0:
            self.mailbox.pending_ins.pop(sender)

    def receive_all(self) -> None:
        """
        Receives mail from all senders.
        """
        for key in list(self.mailbox.pending_ins.keys()):
            self.receive(key)

    @override
    @classmethod
    def convert_from(
        cls,
        obj: Any,
        key: str = "DataFrame",
        **kwargs,
    ) -> "Branch":
        p = cls.get_converter_registry().convert_from(
            target_class=obj,
            key=key,
            **kwargs,
        )
        return cls(messages=p, **kwargs)

    @override
    def convert_to(self, key: str, /, **kwargs: Any) -> Any:
        return self.get_converter_registry().convert_to(
            obj=self,
            key=key,
            **kwargs,
        )

    @property
    def last_response(self) -> AssistantResponse | None:
        """
        Get the last assistant response.

        Returns:
            AssistantResponse | None: The last assistant response, if any.
        """
        for i in reversed(self.progress):
            if isinstance(self.messages[i], AssistantResponse):
                return self.messages[i]

    @property
    def assistant_responses(self) -> Pile:
        """
        Get all assistant responses as a Pile.

        Returns:
            Pile: A Pile containing all assistant responses.
        """
        return self.pile_type(
            [
                self.messages[i]
                for i in self.progress
                if isinstance(self.messages[i], AssistantResponse)
            ]
        )

    def update_last_instruction_meta(self, meta: dict) -> None:
        """
        Update metadata of the last instruction.

        Args:
            meta (dict): Metadata to update.
        """
        for i in reversed(self.progress):
            if isinstance(self.messages[i], Instruction):
                self.messages[i].metadata.update(meta, ["extra"])
                return

    def has_tools(self) -> bool:
        """
        Check if the branch has any registered tools.

        Returns:
            bool: True if tools are registered, False otherwise.
        """
        return self.tool_manager.registry != {}

    def register_tools(self, tools: Any) -> None:
        """
        Register new tools to the tool manager.

        Args:
            tools (Any): Tools to be registered.
        """
        self.tool_manager.register_tools(tools=tools)

    def delete_tools(
        self,
        tools: Any,
        verbose: bool = True,
    ) -> bool:
        """
        Delete specified tools from the tool manager.

        Args:
            tools (Any): Tools to be deleted.
            verbose (bool): If True, print status messages.

        Returns:
            bool: True if deletion was successful, False otherwise.
        """
        if not isinstance(tools, list):
            tools = [tools]
        if is_same_dtype(input_=tools, dtype=str):
            for act_ in tools:
                if act_ in self.tool_manager.registry:
                    self.tool_manager.registry.pop(act_)
            if verbose:
                print("tools successfully deleted")
            return True
        elif is_same_dtype(input_=tools, dtype=Tool):
            for act_ in tools:
                if act_.function_name in self.tool_manager.registry:
                    self.tool_manager.registry.pop(act_.function_name)
            if verbose:
                print("tools successfully deleted")
            return True
        if verbose:
            print("tools deletion failed")
        return False

    def to_chat_messages(self, progress=None) -> list[dict[str, Any]]:
        """
        Convert messages to a list of chat message dictionaries.

        Returns:
            list[dict[str, Any]]: A list of chat message dictionaries.
        """
        if not all(i in self.messages for i in (progress or self.progress)):
            raise ValueError("Invalid progress")
        return [self.messages[i].chat_msg for i in (progress or self.progress)]

    def _is_invoked(self) -> bool:
        """Check if the last message is an ActionResponse."""
        return isinstance(self.messages[-1], ActionResponse)


__all__ = ["Branch"]
# File: lion_core/session/branch.py



from typing import Any, Sequence

from lion_core.libs.data_handlers._to_list import to_list
from lion_core.setting import LN_UNDEFINED


def nset(
    nested_structure: dict[str, Any] | list[Any],
    indices: str | int | Sequence[str | int],
    value: Any,
) -> None:
    """
    Set a value within a nested structure at the specified path defined by indices.

    This method allows setting a value deep within a nested dictionary or list by
    specifying a path to the target location using a sequence of indices. Each index
    in the sequence represents a level in the nested structure, with integers used
    for list indices and strings for dictionary keys.

    Args:
        nested_structure: The nested structure where the value will be set.
        indices: The path of indices leading to where the value should be set.
        value: The value to set at the specified location in the nested structure.

    Raises:
        ValueError: If the indices sequence is empty.
        TypeError: If the target container is not a list or dictionary, or if
                   the index type is incorrect.

    Examples:
        >>> data = {'a': {'b': [10, 20]}}
        >>> nset(data, ['a', 'b', 1], 99)
        >>> assert data == {'a': {'b': [10, 99]}}

        >>> data = [0, [1, 2], 3]
        >>> nset(data, [1, 1], 99)
        >>> assert data == [0, [1, 99], 3]
    """
    if not indices:
        raise ValueError("Indices list is empty, cannot determine target container")

    _indices = to_list(indices)
    target_container = nested_structure

    for i, index in enumerate(_indices[:-1]):
        if isinstance(target_container, list):
            if not isinstance(index, int):
                raise TypeError("Cannot use non-integer index on a list")
            ensure_list_index(target_container, index)
            if target_container[index] is None:
                next_index = _indices[i + 1]
                target_container[index] = [] if isinstance(next_index, int) else {}
        elif isinstance(target_container, dict):
            if isinstance(index, int):
                raise TypeError(
                    f"Unsupported key type: {type(index).__name__}. Only string keys are acceptable."
                )
            if index not in target_container:
                next_index = _indices[i + 1]
                target_container[index] = [] if isinstance(next_index, int) else {}
        else:
            raise TypeError("Target container is not a list or dictionary")

        target_container = target_container[index]

    last_index = _indices[-1]
    if isinstance(target_container, list):
        if not isinstance(last_index, int):
            raise TypeError("Cannot use non-integer index on a list")
        ensure_list_index(target_container, last_index)
        target_container[last_index] = value
    elif isinstance(target_container, dict):
        if isinstance(last_index, int):
            raise TypeError(
                f"Unsupported key type: {type(last_index).__name__}. Only string keys are acceptable."
            )
        target_container[last_index] = value
    else:
        raise TypeError("Cannot set value on non-list/dict element")


def ensure_list_index(lst: list[Any], index: int, default: Any = LN_UNDEFINED) -> None:
    """
    Extend a list to ensure it has a minimum length, appending a default value as needed.

    This utility method ensures that a list is extended to at least a specified
    index plus one. If the list's length is less than this target, it is appended
    with a specified default value until it reaches the required length.

    Args:
        lst: The list to extend.
        index: The target index that the list should reach or exceed.
        default: The value to append to the list for extension. Defaults to None.

    Note:
        Modifies the list in place, ensuring it can safely be indexed at `index`
        without raising an IndexError.
    """
    while len(lst) <= index:
        lst.append(default if default is not LN_UNDEFINED else None)


# Path: lion_core/libs/data_handlers/_nset.py



from lion_core.abc._characteristic import Observable, Temporal
from lion_core.abc._concept import AbstractElement


class BaseRecord(AbstractElement, Observable, Temporal):
    """
    Base class for records. Combines AbstractElement with Observable and
    Temporal characteristics.
    """


class MutableRecord(BaseRecord):
    """
    Mutable record class. Inherits from BaseRecord and allows
    modifications.
    """


class ImmutableRecord(BaseRecord):
    """
    Immutable record class. Inherits from BaseRecord but prevents
    modifications. Once a field is filled with data, that field
    cannot change value.
    """


__all__ = [
    "BaseRecord",
    "MutableRecord",
    "ImmutableRecord",
]

# File: lion_core/abc/record.py



from abc import ABC


class Tao(ABC):
    """
    Foundational abstraction embodying interconnectedness and existence.
    Root for all classes, reflecting Taoist unity and Category Theory relations.
    """


class AbstractSpace(Tao):
    """
    Abstract space or context, aligning with Category Theory's categories.
    Defines domain for elements and interactions, supporting system emergence.
    Subclasses implement __contains__ for membership criteria.
    """


class AbstractElement(Tao):
    """
    Observable entity within a space, reflecting Taoist individuality in unity.
    Embodies Category Theory objects and Complex Systems components.
    Capable of emergent behaviors through interactions.
    """

    @classmethod
    def class_name(cls) -> str:
        """Get class name, supporting reflection and metaprogramming."""
        return cls.__name__


class AbstractObserver(Tao):
    """
    Entity capable of observations, inspired by quantum mechanics and cognition.
    Prepresents intentionality and observer effect in complex systems.
    Subclasses implement specific observation mechanisms.
    """


class AbstractObservation(Tao):
    """
    Act of observing, integrating phenomenology and information theory.
    Captures information exchange and meaning construction in complex systems.
    """


__all__ = [
    "AbstractSpace",
    "AbstractElement",
    "AbstractObserver",
    "AbstractObservation",
]


# File: lion_core/abc/concept.py



import pytest

from lion_core.libs.data_handlers._npop import npop
from lion_core.setting import LN_UNDEFINED


@pytest.mark.parametrize(
    "data, indices, expected_result, expected_data",
    [
        ({"a": {"b": {"c": 3}}}, ["a", "b", "c"], 3, {"a": {"b": {}}}),
        ({"a": {"b": [1, 2, 3]}}, ["a", "b", 2], 3, {"a": {"b": [1, 2]}}),
        ({"a": [1, {"b": 2}]}, ["a", 1, "b"], 2, {"a": [1, {}]}),
        ({"a": 1, "b": 2}, ["b"], 2, {"a": 1}),
        ([1, [2, [3, 4]], 5], [1, 1, 0], 3, [1, [2, [4]], 5]),
        (
            {"key with spaces": {"nested": "value"}},
            ["key with spaces", "nested"],
            "value",
            {"key with spaces": {}},
        ),
        ({0: "zero", 1: "one"}, [1], "one", {0: "zero"}),
        ({"a": {"b": None}}, ["a", "b"], None, {"a": {}}),
        ({"": {"nested": "value"}}, ["", "nested"], "value", {"": {}}),
        ({True: "true", False: "false"}, [True], "true", {False: "false"}),
        ({(1, 2): "tuple key"}, [(1, 2)], "tuple key", {}),
        (
            {"a": {"b": {"c": {"d": 1}}}},
            ["a", "b", "c", "d"],
            1,
            {"a": {"b": {"c": {}}}},
        ),
    ],
)
def test_npop_various_scenarios(data, indices, expected_result, expected_data):
    assert npop(data, indices) == expected_result
    assert data == expected_data


@pytest.mark.parametrize(
    "data, indices, default, expected_result, expected_data",
    [
        ({"a": {"b": 2}}, ["a", "c"], 10, 10, {"a": {"b": 2}}),
        ([], [0], "default", "default", []),
        ({}, ["non_existent"], None, None, {}),
        ({"a": 1}, ["b"], LN_UNDEFINED, LN_UNDEFINED, {"a": 1}),
    ],
)
def test_npop_with_default(data, indices, default, expected_result, expected_data):
    assert npop(data, indices, default=default) == expected_result
    assert data == expected_data


@pytest.mark.parametrize(
    "data, indices",
    [
        ({}, ["a"]),
        ([], [0]),
        ({"a": {"b": 2}}, ["a", "c"]),
        ([1, 2, 3], [3]),
        ({"a": [1, 2]}, ["a", 2]),
        ({"a": {"b": 1}}, ["a", "b", "c"]),
    ],
)
def test_npop_raises_key_error(data, indices):
    with pytest.raises(KeyError):
        npop(data, indices)


@pytest.mark.parametrize(
    "data, indices",
    [
        ([1, 2, 3], ["a"]),
        ({"a": 1}, [0]),
        ({"a": [1, 2, 3]}, ["a", "b"]),
    ],
)
def test_npop_raises_type_error(data, indices):
    with pytest.raises(TypeError):
        npop(data, indices)


def test_npop_empty_indices():
    data = {"a": 1}
    with pytest.raises(ValueError, match="Indices list cannot be empty"):
        npop(data, [])


def test_npop_none_data():
    with pytest.raises(TypeError):
        npop(None, ["a"])


def test_npop_non_subscriptable():
    with pytest.raises(TypeError):
        npop(42, [0])


def test_npop_with_zero_index():
    data = [1, 2, 3]
    assert npop(data, [0]) == 1
    assert data == [2, 3]


def test_npop_with_negative_index():
    data = [1, 2, 3]
    with pytest.raises(ValueError):
        npop(data, [-1])


def test_npop_with_string_index_for_list():
    data = [1, 2, 3]
    with pytest.raises(TypeError):
        npop(data, ["0"])


def test_npop_with_int_index_for_dict():
    data = {"0": "value"}
    assert npop(data, ["0"]) == "value"
    assert data == {}


def test_npop_with_nested_default_dict():
    from collections import defaultdict

    data = defaultdict(lambda: defaultdict(int))
    data["a"]["b"] = 1
    assert npop(data, ["a", "b"]) == 1
    assert dict(data) == {"a": {}}


def test_npop_with_custom_object():
    class CustomObj:
        def __init__(self):
            self.attr = {"key": "value"}

        def __getitem__(self, key):
            return self.attr[key]

        def __delitem__(self, key):
            del self.attr[key]

    obj = CustomObj()
    assert npop(obj, ["key"]) == "value"
    assert obj.attr == {}


def test_npop_with_property():
    class PropObj:
        def __init__(self):
            self._data = {"key": "value"}

        @property
        def data(self):
            return self._data

    obj = PropObj()
    assert npop(obj.data, ["key"]) == "value"
    assert obj._data == {}


@pytest.mark.parametrize(
    "data, indices, expected_result, expected_data",
    [
        ({"a": 1, "b": 2, "c": 3}, ["a", "b", "c"], 3, {"a": 1, "b": 2}),
        ([1, 2, [3, 4, [5, 6]]], [2, 2, 1], 6, [1, 2, [3, 4, [5]]]),
        (
            {"a": [{"b": {"c": [1, 2, 3]}}]},
            ["a", 0, "b", "c", 2],
            3,
            {"a": [{"b": {"c": [1, 2]}}]},
        ),
    ],
)
def test_npop_long_paths(data, indices, expected_result, expected_data):
    assert npop(data, indices) == expected_result
    assert data == expected_data


def test_npop_with_large_nested_structure():
    large_data = {"level1": {}}
    current = large_data["level1"]
    for i in range(2, 101):
        current[f"level{i}"] = {}
        current = current[f"level{i}"]
    current["value"] = "deep"

    indices = [f"level{i}" for i in range(1, 101)] + ["value"]
    assert npop(large_data, indices) == "deep"

    # Verify the structure is empty
    current = large_data["level1"]
    for i in range(2, 101):
        assert current == {}
        if f"level{i}" in current:
            current = current[f"level{i}"]


def test_npop_performance_with_large_list(benchmark):
    large_list = list(range(10**5))

    def pop_last():
        return npop(large_list, [-1])

    result = benchmark(pop_last)
    assert result == 99999
    assert len(large_list) == 99999


def test_npop_with_all_python_basic_types():
    data = {
        "int": 1,
        "float": 2.0,
        "complex": 1 + 2j,
        "str": "string",
        "list": [1, 2, 3],
        "tuple": (4, 5, 6),
        "dict": {"key": "value"},
        "set": {7, 8, 9},
        "frozenset": frozenset([10, 11, 12]),
        "bool": True,
        "none": None,
        "bytes": b"bytes",
        "bytearray": bytearray(b"bytearray"),
    }
    for key in list(data.keys()):
        assert npop(data, [key]) == data[key]
    assert data == {}


@pytest.mark.parametrize(
    "data, indices, expected_result, expected_data",
    [
        ({"a": {"b": {"c": 3}}}, "a.b.c", 3, {"a": {"b": {}}}),
        ([1, [2, [3, 4]]], "1.1.0", 3, [1, [2, [4]]]),
        ({"a": [1, {"b": 2}]}, "a.1.b", 2, {"a": [1, {}]}),
    ],
)
def test_npop_with_string_indices(data, indices, expected_result, expected_data):
    assert npop(data, indices.split(".")) == expected_result
    assert data == expected_data


def test_npop_with_generator():
    def gen():
        yield 1
        yield 2
        yield 3

    data = {"gen": gen()}
    with pytest.raises(TypeError):
        npop(data, ["gen", 1])


def test_npop_with_recursive_structure():
    data = {"a": 1}
    data["b"] = data
    assert npop(data, ["a"]) == 1
    assert data == {"b": {"b": {...}}}


def test_npop_with_custom_classes():
    class CustomList(list):
        pass

    class CustomDict(dict):
        pass

    data = CustomDict({"a": CustomList([1, 2, 3])})
    assert npop(data, ["a", 1]) == 2
    assert data == CustomDict({"a": CustomList([1, 3])})


# File: tests/test_npop.py



from lion_core.libs.algorithms.jaro_distance import (
    jaro_distance,
    jaro_winkler_similarity,
)
from lion_core.libs.algorithms.levenshtein_distance import levenshtein_distance

__all__ = ["jaro_distance", "jaro_winkler_similarity", "levenshtein_distance"]



"""Provide utility for forcefully validating and converting inputs to boolean."""

from typing import Any

from lion_core.libs.data_handlers import strip_lower


def validate_boolean(x: Any) -> bool:
    """
    Forcefully validate and convert the input into a boolean value.

    This function attempts to convert various input types to a boolean value.
    It recognizes common string representations of true and false, as well
    as numeric values.

    Args:
        x: The input to be converted to boolean.

    Returns:
        The boolean representation of the input.

    Raises:
        ValueError: If the input cannot be converted to a boolean value.

    Examples:
        >>> force_validate_boolean("true")
        True
        >>> force_validate_boolean("false")
        False
        >>> force_validate_boolean("yes")
        True
        >>> force_validate_boolean("no")
        False
        >>> force_validate_boolean(True)
        True
        >>> force_validate_boolean("1")
        True
        >>> force_validate_boolean("0")
        False
    """
    if isinstance(x, bool):
        return x

    if strip_lower(x) in ["true", "1", "correct", "yes"]:
        return True

    elif strip_lower(x) in ["false", "0", "incorrect", "no", "none", "n/a"]:
        return False

    raise ValueError(f"Failed to convert {x} into a boolean value")


# File: lion_core/libs/parsers/_force_validate_boolean.py



import asyncio
from typing import Any, Callable, Sequence, TypeVar

from lion_core.libs.function_handlers._ucall import ucall
from lion_core.libs.function_handlers._util import is_coroutine_func
from lion_core.setting import LN_UNDEFINED

T = TypeVar("T")
ErrorHandler = Callable[[Exception], Any]


async def pcall(
    funcs: Sequence[Callable[..., T]],
    retries: int = 0,
    initial_delay: float = 0,
    delay: float = 0,
    backoff_factor: float = 1,
    default: Any = LN_UNDEFINED,
    timeout: float | None = None,
    timing: bool = False,
    verbose: bool = True,
    error_msg: str | None = None,
    error_map: dict[type, ErrorHandler] | None = None,
    max_concurrent: int | None = None,
    throttle_period: float | None = None,
    **kwargs: Any,
) -> list[T] | list[tuple[T, float]]:
    """
    Execute multiple functions asynchronously with customizable options.

    Args:
        funcs: Sequence of functions to be executed.
        retries: Number of retry attempts for each function.
        initial_delay: Initial delay before starting the execution.
        delay: Delay between retry attempts.
        backoff_factor: Factor by which delay increases after each attempt.
        default: Default value to return if all attempts fail.
        timeout: Timeout for each function execution.
        timing: Whether to return the execution duration.
        verbose: Whether to print retry messages.
        error_msg: Custom error message.
        error_map: Dictionary mapping exception types to error handlers.
        max_concurrent: Maximum number of concurrent executions.
        throttle_period: Minimum time period between function executions.
        **kwargs: Additional keyword arguments for each function.

    Returns:
        List of results, optionally including execution durations if timing
        is True.
    """
    if initial_delay:
        await asyncio.sleep(initial_delay)

    semaphore = asyncio.Semaphore(max_concurrent) if max_concurrent else None
    throttle_delay = throttle_period if throttle_period else 0

    async def _task(func: Callable[..., Any], index: int) -> Any:
        if semaphore:
            async with semaphore:
                return await _execute_task(func, index)
        else:
            return await _execute_task(func, index)

    async def _execute_task(func: Callable[..., Any], index: int) -> Any:
        attempts = 0
        current_delay = delay
        while True:
            try:
                if timing:
                    start_time = asyncio.get_event_loop().time()
                    result = await asyncio.wait_for(ucall(func, **kwargs), timeout)
                    end_time = asyncio.get_event_loop().time()
                    return index, result, end_time - start_time
                else:
                    result = await asyncio.wait_for(ucall(func, **kwargs), timeout)
                    return index, result
            except asyncio.TimeoutError as e:
                raise asyncio.TimeoutError(
                    f"{error_msg or ''} Timeout {timeout} seconds exceeded"
                ) from e
            except Exception as e:
                if error_map and type(e) in error_map:
                    handler = error_map[type(e)]
                    if is_coroutine_func(handler):
                        return index, await handler(e)
                    else:
                        return index, handler(e)
                attempts += 1
                if attempts <= retries:
                    if verbose:
                        print(
                            f"Attempt {attempts}/{retries + 1} failed: {e}, retrying..."
                        )
                    await asyncio.sleep(current_delay)
                    current_delay *= backoff_factor
                else:
                    if default is not LN_UNDEFINED:
                        return index, default
                    raise e

    tasks = [_task(func, index) for index, func in enumerate(funcs)]
    results = []
    for coro in asyncio.as_completed(tasks):
        result = await coro
        results.append(result)
        await asyncio.sleep(throttle_delay)

    results.sort(key=lambda x: x[0])  # Sort results based on the original index

    if timing:
        return [(result[1], result[2]) for result in results]
    else:
        return [result[1] for result in results]


# Path: lion_core/libs/function_handlers/_pcall.py



import pytest

from lion_core.action.function_calling import FunctionCalling
from lion_core.action.tool import Tool, func_to_tool
from lion_core.action.tool_manager import ToolManager


# Mock functions for testing
def mock_function(x: int, y: int) -> int:
    return x + y


async def async_mock_function(x: int, y: int) -> int:
    return x + y


# Fixtures
@pytest.fixture
def tool_manager():
    return ToolManager()


@pytest.fixture
def sample_tool():
    return Tool(function=mock_function)


@pytest.fixture
def sample_async_tool():
    return Tool(function=async_mock_function)


# Test initialization
def test_init():
    tm = ToolManager()
    assert isinstance(tm.registry, dict)
    assert len(tm.registry) == 0


# Test __contains__
def test_contains(tool_manager, sample_tool):
    tool_manager.register_tool(sample_tool)
    assert sample_tool in tool_manager
    assert "mock_function" in tool_manager
    assert mock_function in tool_manager
    assert "non_existent_tool" not in tool_manager


# Test register_tool
def test_register_tool(tool_manager, sample_tool):
    tool_manager.register_tool(sample_tool)
    assert "mock_function" in tool_manager.registry

    with pytest.raises(ValueError):
        tool_manager.register_tool(sample_tool)

    tool_manager.register_tool(sample_tool, update=True)
    assert tool_manager.registry["mock_function"] == sample_tool


def test_register_callable(tool_manager):
    tool_manager.register_tool(mock_function)
    assert "mock_function" in tool_manager.registry
    assert isinstance(tool_manager.registry["mock_function"], Tool)


def test_register_invalid_tool(tool_manager):
    with pytest.raises(TypeError):
        tool_manager.register_tool("not_a_tool")


# Test match_tool
def test_match_tool_tuple(tool_manager, sample_tool):
    tool_manager.register_tool(sample_tool)
    func_call = ("mock_function", {"x": 1, "y": 2})
    result = tool_manager.match_tool(func_call)
    assert isinstance(result, FunctionCalling)
    assert result.func_tool == sample_tool
    assert result.arguments == {"x": 1, "y": 2}


def test_match_tool_dict(tool_manager, sample_tool):
    tool_manager.register_tool(sample_tool)
    func_call = {"function": "mock_function", "arguments": {"x": 1, "y": 2}}
    result = tool_manager.match_tool(func_call)
    assert isinstance(result, FunctionCalling)
    assert result.func_tool == sample_tool
    assert result.arguments == {"x": 1, "y": 2}


def test_match_tool_str(tool_manager, sample_tool):
    tool_manager.register_tool(sample_tool)
    func_call = '{"function": "mock_function", "arguments": {"x": 1, "y": 2}}'
    result = tool_manager.match_tool(func_call)
    assert isinstance(result, FunctionCalling)
    assert result.func_tool == sample_tool
    assert result.arguments == {"x": 1, "y": 2}


def test_match_tool_invalid(tool_manager):
    with pytest.raises(TypeError):
        tool_manager.match_tool(123)

    with pytest.raises(ValueError):
        tool_manager.match_tool(("non_existent_function", {}))

    with pytest.raises(ValueError):
        tool_manager.match_tool({"invalid": "format"})

    with pytest.raises(ValueError):
        tool_manager.match_tool("invalid_json_string")


# Test invoke
@pytest.mark.asyncio
async def test_invoke(tool_manager, sample_tool):
    tool_manager.register_tool(sample_tool)
    result = await tool_manager.invoke(("mock_function", {"x": 1, "y": 2}))
    assert result == 3


@pytest.mark.asyncio
async def test_invoke_async(tool_manager, sample_async_tool):
    tool_manager.register_tool(sample_async_tool)
    result = await tool_manager.invoke(("async_mock_function", {"x": 1, "y": 2}))
    assert result == 3


# Edge cases and additional tests
def test_empty_tool_manager(tool_manager):
    assert len(tool_manager.registry) == 0
    assert tool_manager.schema_list == []
    with pytest.raises(ValueError):
        tool_manager.match_tool(("any_function", {}))


def test_register_tool_with_same_name(tool_manager):
    def func1():
        pass

    def func2():
        pass

    tool1 = func_to_tool(func1)[0]
    tool2 = func_to_tool(func2)[0]

    tool_manager.register_tool(tool1)
    tool_manager.register_tool(tool2)  # This should update the existing tool
    assert tool_manager.registry[tool2.function_name] == tool2


def test_match_tool_with_extra_args(tool_manager, sample_tool):
    tool_manager.register_tool(sample_tool)
    func_call = ("mock_function", {"x": 1, "y": 2, "z": 3})
    result = tool_manager.match_tool(func_call)
    assert result.arguments == {"x": 1, "y": 2, "z": 3}


def test_get_tool_schema_with_additional_kwargs(tool_manager, sample_tool):
    tool_manager.register_tool(sample_tool)
    result = tool_manager.get_tool_schema(sample_tool, extra_param="test")
    assert "tools" in result
    assert isinstance(result["tools"], dict)
    assert result["extra_param"] == "test"


print("All tests for ToolManager completed successfully!")



from lion_core.communication.action_request import ActionRequest
from lion_core.communication.action_response import ActionResponse


def handle_action_request(
    *,
    sender,
    recipient,
    action_request,
    func,
    arguments,
):
    if action_request:
        if not isinstance(action_request, ActionRequest):
            raise ValueError(
                "Error: action request must be an instance of ActionRequest."
            )
        return action_request

    if func:
        if callable(func):
            func = func.__name__
        if not arguments:
            raise ValueError("Error: please provide arguments for the function.")
        return ActionRequest(
            func=func,
            arguments=arguments,
            sender=sender,
            recipient=recipient,
        )


def handle_action_response(
    *,
    sender,
    action_request,
    action_response,
    func_output,
):
    if func_output or action_response:
        if not action_request or not isinstance(action_request, ActionRequest):
            raise ValueError(
                "Error: please provide a corresponding action request for an "
                "action response."
            )

        # if action response is provided, we update the request with the response
        # and return the response
        if isinstance(action_response, ActionResponse):
            action_response.update_request(
                action_request=action_request,
                func_output=func_output,
            )
            return action_response

        # if only function output is provided, we create a new action response
        return ActionResponse(
            action_request=action_request,
            sender=sender,
            func_outputs=func_output,
        )

    if action_request:
        if not isinstance(action_request, ActionRequest):
            raise ValueError(
                "Error: action request must be an instance of ActionRequest."
            )
        return action_request


def handle_action(
    *,
    sender,
    recipient,
    action_request,
    action_response,
    func,
    arguments,
    func_output,
) -> ActionRequest | ActionResponse | None:
    a = handle_action_response(
        sender=sender,
        action_request=action_request,
        action_response=action_response,
        func_output=func_output,
    )
    if isinstance(a, ActionResponse):
        return a

    return handle_action_request(
        sender=sender,
        recipient=recipient,
        action_request=action_request,
        func=func,
        arguments=arguments,
    )



from collections import OrderedDict, deque, namedtuple

import pytest
from pydantic import BaseModel

from lion_core.libs.data_handlers._to_list import to_list
from lion_core.setting import LionUndefined


class CustomIterable:
    def __iter__(self):
        return iter([1, 2, 3])


class CustomMapping:
    def __getitem__(self, key):
        return {1: "a", 2: "b", 3: "c"}[key]

    def keys(self):
        return [1, 2, 3]


class TestModel(BaseModel):
    field: str = "value"


@pytest.mark.parametrize(
    "input_value, expected",
    [
        (None, []),
        (LionUndefined(), []),
    ],
)
def test_none_and_undefined_input(input_value, expected):
    assert to_list(input_value) == expected


@pytest.mark.parametrize(
    "input_value, expected",
    [
        (1, [1]),
        (1.5, [1.5]),
        (True, [True]),
        ("string", ["string"]),
        (b"bytes", [b"bytes"]),
        (bytearray(b"bytearray"), [bytearray(b"bytearray")]),
    ],
)
def test_primitive_types(input_value, expected):
    assert to_list(input_value) == expected


@pytest.mark.parametrize(
    "input_value, expected, flatten",
    [
        ([1, 2, 3], [1, 2, 3], False),
        ([1, [2, 3]], [1, [2, 3]], False),
        ([1, [2, 3]], [1, 2, 3], True),
    ],
)
def test_list_input(input_value, expected, flatten):
    assert to_list(input_value, flatten=flatten) == expected


@pytest.mark.parametrize(
    "input_value, expected, flatten",
    [
        ((1, 2, 3), [1, 2, 3], False),
        ((1, (2, 3)), [1, (2, 3)], False),
        ((1, (2, 3)), [1, 2, 3], True),
    ],
)
def test_tuple_input(input_value, expected, flatten):
    assert to_list(input_value, flatten=flatten) == expected


def test_set_input():
    assert set(to_list({1, 2, 3})) == {1, 2, 3}


def test_dict_input():
    assert to_list({"a": 1, "b": 2}) == [{"a": 1, "b": 2}]


def test_ordereddict_input():
    od = OrderedDict([("a", 1), ("b", 2)])
    assert to_list(od) == [od]


def test_custom_mapping():
    cm = CustomMapping()
    assert to_list(cm) == [cm]


def test_generator_input():
    def gen():
        yield from range(3)

    assert to_list(gen()) == [0, 1, 2]


def test_custom_iterable():
    ci = CustomIterable()
    assert to_list(ci) == [1, 2, 3]


def test_deque_input():
    d = deque([1, 2, 3])
    assert to_list(d) == [1, 2, 3]


def test_namedtuple_input():
    Point = namedtuple("Point", ["x", "y"])
    p = Point(1, 2)
    assert to_list(p) == [1, 2]


def test_pydantic_model():
    model = TestModel()
    assert to_list(model) == [model]


@pytest.mark.parametrize(
    "input_value, expected, flatten",
    [
        ([1, [2, [3, [4]]]], [1, [2, [3, [4]]]], False),
        ([1, [2, [3, [4]]]], [1, 2, 3, 4], True),
    ],
)
def test_nested_structures(input_value, expected, flatten):
    assert to_list(input_value, flatten=flatten) == expected


@pytest.mark.parametrize(
    "input_value, expected, flatten",
    [
        ([1, "two", [3, [4, "five"]]], [1, "two", [3, [4, "five"]]], False),
        ([1, "two", [3, [4, "five"]]], [1, "two", 3, 4, "five"], True),
    ],
)
def test_mixed_types(input_value, expected, flatten):
    assert to_list(input_value, flatten=flatten) == expected


@pytest.mark.parametrize(
    "input_value, expected, flatten, dropna",
    [
        ([1, None, 2, [3, None, 4]], [1, 2, [3, 4]], False, True),
        ([1, None, 2, [3, None, 4]], [1, 2, 3, 4], True, True),
    ],
)
def test_dropna(input_value, expected, flatten, dropna):
    assert to_list(input_value, flatten=flatten, dropna=dropna) == expected


@pytest.mark.parametrize(
    "input_value, expected",
    [
        ([], []),
        ({}, [{}]),
        (set(), []),
    ],
)
def test_empty_inputs(input_value, expected):
    assert to_list(input_value) == expected


def test_large_input():
    large_list = list(range(10000))
    assert to_list(large_list) == large_list


@pytest.mark.parametrize(
    "input_value, expected",
    [
        ([1, "nested", [2, ["deep", "list"]]], [1, "nested", 2, "deep", "list"]),
    ],
)
def test_flatten_with_strings(input_value, expected):
    assert to_list(input_value, flatten=True) == expected


@pytest.mark.parametrize(
    "input_value, expected",
    [
        ([1, (2, 3), [4, (5, 6)]], [1, 2, 3, 4, 5, 6]),
    ],
)
def test_flatten_with_tuple(input_value, expected):
    assert to_list(input_value, flatten=True) == expected


def test_flatten_with_set():
    input_with_set = [1, {2, 3}, [4, {5, 6}]]
    flattened = to_list(input_with_set, flatten=True)
    assert len(flattened) == 6
    assert set(flattened) == {1, 2, 3, 4, 5, 6}


def test_flatten_with_dict():
    input_with_dict = [1, {"a": 2}, [3, {"b": 4}]]
    expected = [1, {"a": 2}, 3, {"b": 4}]
    assert to_list(input_with_dict, flatten=True) == expected


def test_generator_with_none():
    def gen_with_none():
        yield 1
        yield None
        yield 2

    assert to_list(gen_with_none(), dropna=True) == [1, 2]


def test_deeply_nested_structure():
    deeply_nested = [1, [2, [3, [4, [5, [6]]]]]]
    assert to_list(deeply_nested, flatten=True) == [1, 2, 3, 4, 5, 6]


def test_input_with_all_none():
    assert to_list([None, None, None], dropna=True) == []


def test_custom_objects():
    class CustomObj:
        pass

    obj = CustomObj()
    assert to_list(obj) == [obj]


def test_flatten_with_empty_nested_structures():
    input_with_empty = [1, [], [2, []], [[], 3]]
    assert to_list(input_with_empty, flatten=True) == [1, 2, 3]


def test_flatten_and_dropna_combination():
    complex_input = [1, None, [2, None, [3, None, 4]]]
    assert to_list(complex_input, flatten=True, dropna=True) == [1, 2, 3, 4]


def test_input_types_preservation():
    mixed_types = [1, "two", 3.0, True, b"four"]
    assert to_list(mixed_types) == mixed_types


def test_very_large_nested_structure():
    large_nested = list(range(1000)) + [list(range(1000, 2000))]
    flattened = to_list(large_nested, flatten=True)
    assert len(flattened) == 2000
    assert flattened[-1] == 1999


def test_to_list_with_frozenset():
    fs = frozenset([1, 2, 3])
    assert set(to_list(fs)) == {1, 2, 3}


def test_to_list_with_view():
    d = {1: "a", 2: "b", 3: "c"}
    assert set(to_list(d.keys())) == {1, 2, 3}
    assert set(to_list(d.values())) == {"a", "b", "c"}


def test_to_list_with_enum():
    from enum import Enum

    class Color(Enum):
        RED = 1
        GREEN = 2
        BLUE = 3

    assert to_list(Color) == [Color.RED, Color.GREEN, Color.BLUE]


def test_to_list_with_range():
    r = range(5)
    assert to_list(r) == [0, 1, 2, 3, 4]


def test_to_list_with_iterator():
    it = iter([1, 2, 3])
    assert to_list(it) == [1, 2, 3]


def test_to_list_with_generator_expression():
    gen_exp = (x for x in range(5))
    assert to_list(gen_exp) == [0, 1, 2, 3, 4]


def test_to_list_with_zip():
    z = zip([1, 2, 3], ["a", "b", "c"])
    assert to_list(z) == [(1, "a"), (2, "b"), (3, "c")]


def test_to_list_with_filter():
    f = filter(lambda x: x % 2 == 0, range(10))
    assert to_list(f) == [0, 2, 4, 6, 8]


def test_to_list_with_map():
    m = map(lambda x: x * 2, range(5))
    assert to_list(m) == [0, 2, 4, 6, 8]


def test_to_list_with_bytes():
    b = b"hello"
    assert to_list(b) == [104, 101, 108, 108, 111]


def test_to_list_with_memoryview():
    mv = memoryview(b"hello")
    assert to_list(mv) == [104, 101, 108, 108, 111]


def test_to_list_with_complex():
    c = 1 + 2j
    assert to_list(c) == [c]


def test_to_list_performance_large_nested(benchmark):
    def create_large_nested():
        return [list(range(1000)) for _ in range(1000)]

    large_nested = create_large_nested()
    result = benchmark(to_list, large_nested, flatten=True)
    assert len(result) == 1_000_000
    assert result[0] == 0
    assert result[-1] == 999


def test_to_list_with_custom_flatten():
    class CustomFlatten:
        def __iter__(self):
            return iter([1, [2, 3]])

    cf = CustomFlatten()
    assert to_list(cf, flatten=True) == [1, 2, 3]


# File: tests/test_to_list.py



from lion_core.libs.parsers._as_readable_json import as_readable_json
from lion_core.libs.parsers._choose_most_similar import choose_most_similar
from lion_core.libs.parsers._extract_code_block import extract_code_block
from lion_core.libs.parsers._extract_docstring import extract_docstring_details
from lion_core.libs.parsers._fuzzy_parse_json import fix_json_string, fuzzy_parse_json
from lion_core.libs.parsers._md_to_json import (
    escape_chars_in_json,
    extract_json_block,
    md_to_json,
)

from ._function_to_schema import function_to_schema
from ._validate_boolean import validate_boolean
from ._validate_keys import validate_keys
from ._validate_mapping import validate_mapping
from ._xml_parser import xml_to_dict

__all__ = [
    "choose_most_similar",
    "extract_docstring_details",
    "extract_code_block",
    "md_to_json",
    "as_readable_json",
    "fuzzy_parse_json",
    "validate_mapping",
    "validate_keys",
    "validate_boolean",
    "function_to_schema",
    "extract_json_block",
    "fix_json_string",
    "escape_chars_in_json",
    "xml_to_dict",
]



"""Provide utility for converting various inputs to readable JSON strings."""

from typing import Any

from lion_core.libs.data_handlers import to_dict, to_str


def as_readable_json(input_: Any, /, **kwargs) -> str:
    """
    Convert the input to a readable JSON string.

    This function attempts to convert the input to a dictionary and then
    to a formatted JSON string. If conversion to a dictionary fails, it
    returns the string representation of the input.

    Args:
        input_: The input to be converted to a readable JSON string.
        kwargs for to_dict

    Returns:
        A formatted JSON string if the input can be converted to a
        dictionary, otherwise the string representation of the input.

    Raises:
        ValueError: If the input cannot be converted to a readable dict.
    """
    try:
        dict_ = to_dict(input_, **kwargs)
        config = {"indent": 4} if isinstance(dict_, dict) else {}
        return to_str(dict_, **config)
    except Exception as e:
        raise ValueError(f"Could not convert given input to readable dict: {e}") from e


# Path: lion_core/libs/parsers/_as_readable_json.py



import unittest

from lion_core.libs.parsers import *
from lion_core.libs.parsers._extract_docstring import (
    _extract_docstring_details_google,
    _extract_docstring_details_rest,
)
from lion_core.libs.parsers._md_to_json import escape_chars_in_json


class TestChooseMostSimilar(unittest.TestCase):
    def test_choose_most_similar_default(self):
        word = "helo"
        correct_words_list = ["hello", "help", "hero", "hill"]
        result = choose_most_similar(word, correct_words_list)
        self.assertEqual(result, "hello")

    def test_choose_most_similar_custom_func(self):
        def dummy_similarity(a, b):
            return 1.0 if a == b else 0.0

        word = "test"
        correct_words_list = ["test", "toast", "taste"]
        result = choose_most_similar(
            word, correct_words_list, score_func=dummy_similarity
        )
        self.assertEqual(result, "test")

    def test_choose_most_similar_empty_list(self):
        word = "test"
        correct_words_list = []
        result = choose_most_similar(word, correct_words_list)
        self.assertIsNone(result)

    def test_choose_most_similar_no_score_func(self):
        word = "world"
        correct_words_list = ["word", "worn", "wild"]
        result = choose_most_similar(word, correct_words_list)
        self.assertIn(
            result, correct_words_list
        )  # Since the default function is used, any close match is acceptable.


class TestExtractCodeBlocks(unittest.TestCase):
    def test_extract_python_code(self):
        input_text = (
            "Here is some Python code:\n```python\nprint('Hello, world!')\n```\n"
        )
        expected_output = "print('Hello, world!')"
        self.assertEqual(extract_code_block(input_text), expected_output)

    def test_extract_java_code(self):
        input_text = "Here is some Java code:\n```java\nSystem.out.println('Hello, world!');\n```\n"
        expected_output = "System.out.println('Hello, world!');"
        self.assertEqual(extract_code_block(input_text), expected_output)

    def test_extract_javascript_code(self):
        input_text = "Here is some JavaScript code:\n```javascript\nconsole.log('Hello, world!');\n```\n"
        expected_output = "console.log('Hello, world!');"
        self.assertEqual(extract_code_block(input_text), expected_output)

    def test_extract_cpp_code(self):
        input_text = "Here is some C++ code:\n```cpp\n#include <iostream>\nint main() { std::cout << 'Hello, world!'; return 0; }\n```\n"
        expected_output = "#include <iostream>\nint main() { std::cout << 'Hello, world!'; return 0; }"
        self.assertEqual(extract_code_block(input_text), expected_output)

    def test_extract_ruby_code(self):
        input_text = "Here is some Ruby code:\n```ruby\nputs 'Hello, world!'\n```\n"
        expected_output = "puts 'Hello, world!'"
        self.assertEqual(extract_code_block(input_text), expected_output)


class TestExtractDocstringDetails(unittest.TestCase):
    def test_extract_google_style(self):
        def sample_func():
            """
            Sample function.

            Args:
                    param1 (int): Description of param1.
                    param2 (str): Description of param2.
            """
            pass

        description, params = _extract_docstring_details_google(sample_func)
        self.assertEqual(description, "Sample function.")
        self.assertDictEqual(
            params,
            {"param1": "Description of param1.", "param2": "Description of param2."},
        )

    def test_extract_reST_style(self):
        def sample_func():
            """
            Sample function.

            :param param1: Description of param1.
            :type param1: int
            :param param2: Description of param2.
            :type param2: str
            """
            pass

        description, params = _extract_docstring_details_rest(sample_func)
        self.assertEqual(description, "Sample function.")
        self.assertDictEqual(
            params,
            {"param1": "Description of param1.", "param2": "Description of param2."},
        )

    def test_google_style(self):
        def example_function(param1: int, param2: str):
            """
            Example function.

            Args:
                param1 (int): The first parameter.
                param2 (str): The second parameter.
            """
            pass

        description, params = extract_docstring_details(
            example_function, style="google"
        )
        self.assertEqual(description, "Example function.")
        self.assertEqual(
            params,
            {"param1": "The first parameter.", "param2": "The second parameter."},
        )

    def test_rest_style(self):
        def example_function(param1: int, param2: str):
            """
            Example function.

            :param param1: The first parameter.
            :type param1: int
            :param param2: The second parameter.
            :type param2: str
            """
            pass

        description, params = extract_docstring_details(example_function, style="rest")
        self.assertEqual(description, "Example function.")
        self.assertEqual(
            params,
            {"param1": "The first parameter.", "param2": "The second parameter."},
        )

    def test_no_docstring(self):
        def example_function(param1: int, param2: str):
            pass

        description, params = extract_docstring_details(
            example_function, style="google"
        )
        self.assertEqual(description, None)
        self.assertEqual(params, {})

    def test_unsupported_style(self):
        def example_function(param1: int, param2: str):
            """
            Example function.

            Args:
                param1 (int): The first parameter.
                param2 (str): The second parameter.
            """
            pass

        with self.assertRaises(ValueError):
            extract_docstring_details(example_function, style="unsupported")

    def test_google_style_incomplete(self):
        def example_function(param1: int, param2: str):
            """
            Example function.

            Args:
                param1: The first parameter.
                param2 (str):
            """
            pass

        description, params = extract_docstring_details(
            example_function, style="google"
        )
        self.assertEqual(description, "Example function.")
        self.assertEqual(params, {"param1": "The first parameter.", "param2": ""})


class TestForceValidateBoolean(unittest.TestCase):
    def test_boolean_true(self):
        self.assertTrue(validate_boolean(True))
        self.assertTrue(validate_boolean("true"))
        self.assertTrue(validate_boolean("1"))
        self.assertTrue(validate_boolean("yes"))
        self.assertTrue(validate_boolean("correct"))

    def test_boolean_false(self):
        self.assertFalse(validate_boolean(False))
        self.assertFalse(validate_boolean("false"))
        self.assertFalse(validate_boolean("0"))
        self.assertFalse(validate_boolean("no"))
        self.assertFalse(validate_boolean("incorrect"))
        self.assertFalse(validate_boolean("none"))
        self.assertFalse(validate_boolean("n/a"))

    def test_invalid_conversion(self):
        try:
            validate_boolean("invalid")
        except Exception as e:
            self.assertIsInstance(e, ValueError)

        try:
            validate_boolean(123)
        except Exception as e:
            self.assertIsInstance(e, ValueError)

        try:
            validate_boolean(None)
        except Exception as e:
            self.assertIsInstance(e, ValueError)

    def test_strip_whitespace(self):
        self.assertTrue(validate_boolean("  true  "))
        self.assertFalse(validate_boolean("  false  "))

    def test_case_insensitive(self):
        self.assertTrue(validate_boolean("TRUE"))
        self.assertTrue(validate_boolean("Yes"))
        self.assertFalse(validate_boolean("FALSE"))
        self.assertFalse(validate_boolean("No"))


class TestForceValidateKeys2(unittest.TestCase):
    def test_exact_match(self):
        keys = ["name", "age", "location"]
        input_dict = {"name": "John", "age": 30, "location": "New York"}
        expected_output = {"name": "John", "age": 30, "location": "New York"}
        self.assertEqual(validate_keys(input_dict, keys), expected_output)

    def test_partial_match(self):
        keys = ["name", "age", "location"]
        input_dict = {"nme": "John", "ag": 30, "loc": "New York"}
        expected_output = {"name": "John", "age": 30, "location": "New York"}
        self.assertEqual(validate_keys(input_dict, keys), expected_output)

    def test_additional_keys(self):
        keys = ["name", "age"]
        input_dict = {"name": "John", "age": 30, "location": "New York"}
        expected_output = {"name": "John", "age": 30, "location": "New York"}
        self.assertEqual(validate_keys(input_dict, keys), expected_output)

    def test_missing_keys(self):
        keys = ["name", "age", "location"]
        input_dict = {"name": "John"}
        expected_output = {"name": "John"}
        self.assertEqual(validate_keys(input_dict, keys), expected_output)

    def test_custom_score_func(self):
        def dummy_similarity(a, b):
            return 1.0 if a == b else 0.0

        keys = ["name", "age", "location"]
        input_dict = {"name": "John", "age": 30, "loc": "New York"}
        expected_output = {"name": "John", "age": 30, "location": "New York"}
        self.assertEqual(
            validate_keys(input_dict, keys, score_func=dummy_similarity),
            expected_output,
        )


class TestForceValidateKeys(unittest.TestCase):
    def test_match_mode(self):
        keys = ["name", "age", "location"]
        dict_ = {"name": "John", "age": 30, "loc": "NYC"}
        result = validate_keys(dict_, keys)
        self.assertEqual(result, {"name": "John", "age": 30, "location": "NYC"})

    def test_superset_mode(self):
        keys = ["name", "age"]
        dict_ = {"name": "John", "age": 30, "location": "NYC"}
        result = validate_keys(dict_, keys)
        self.assertEqual(result, {"name": "John", "age": 30, "location": "NYC"})

    def test_subset_mode(self):
        keys = ["name", "age", "location"]
        dict_ = {"name": "John", "age": 30}
        result = validate_keys(dict_, keys)
        self.assertEqual(result, {"name": "John", "age": 30})

    def test_handle_unmatched_ignore(self):
        keys = ["name", "age"]
        dict_ = {"name": "John", "age": 30, "location": "NYC"}
        result = validate_keys(dict_, keys, handle_unmatched="ignore")
        self.assertEqual(result, {"name": "John", "age": 30, "location": "NYC"})

    def test_handle_unmatched_force(self):
        keys = ["name", "age", "location"]
        dict_ = {"name": "John", "age": 30}
        result = validate_keys(
            dict_, keys, handle_unmatched="force", fill_value="Unknown"
        )
        self.assertEqual(result, {"name": "John", "age": 30, "location": "Unknown"})

    def test_handle_unmatched_remove(self):
        keys = ["name", "age"]
        dict_ = {"name": "John", "age": 30, "location": "NYC"}
        result = validate_keys(dict_, keys, handle_unmatched="remove")
        self.assertEqual(result, {"name": "John", "age": 30})

    def test_handle_unmatched_raise(self):
        keys = ["name", "age", "location"]
        dict_ = {"name": "John", "age": 30}
        with self.assertRaises(ValueError):
            validate_keys(dict_, keys, handle_unmatched="raise")

    def test_handle_unmatched_fill(self):
        keys = ["name", "age", "location"]
        dict_ = {"name": "John", "age": 30}
        result = validate_keys(
            dict_, keys, handle_unmatched="fill", fill_value="Unknown"
        )
        self.assertEqual(result, {"name": "John", "age": 30, "location": "Unknown"})

    def test_empty_dict(self):
        keys = ["name", "age", "location"]
        dict_ = {}
        result = validate_keys(
            dict_, keys, handle_unmatched="fill", fill_value="Unknown"
        )
        self.assertEqual(
            result, {"name": "Unknown", "age": "Unknown", "location": "Unknown"}
        )

    def test_empty_keys(self):
        keys = []
        dict_ = {"name": "John", "age": 30, "location": "NYC"}
        result = validate_keys(dict_, keys)
        self.assertEqual(result, {"name": "John", "age": 30, "location": "NYC"})

    def test_custom_score_func(self):
        def custom_score_func(a, b):
            return len(set(a) & set(b)) / len(set(a) | set(b))

        keys = ["name", "age", "location"]
        dict_ = {"name": "John", "age": 30, "loc": "NYC"}
        result = validate_keys(dict_, keys, score_func=custom_score_func)
        self.assertEqual(result, {"name": "John", "age": 30, "location": "NYC"})

    def test_no_matching_keys(self):
        keys = ["gender", "nationality"]
        dict_ = {"name": "John", "age": 30, "location": "NYC"}

        try:
            result = validate_keys(
                dict_, keys, handle_unmatched="force", fill_value="Unknown", strict=True
            )
            self.assertEqual(
                result, {"gender": 30, "nationality": "John", "location": "NYC"}
            )
        except Exception as e:
            self.assertIsInstance(e, ValueError)

    def test_keys_as_dict(self):
        keys = {"name": str, "age": int, "location": str}
        dict_ = {"name": "John", "age": 30, "loc": "NYC"}
        result = validate_keys(dict_, keys)
        self.assertEqual(result, {"name": "John", "age": 30, "location": "NYC"})

    def test_fill_value_with_different_types(self):
        keys = ["name", "age", "is_student"]
        dict_ = {"name": "John"}
        fill_mapping = {"age": 0, "is_student": False}
        result = validate_keys(
            dict_, keys, handle_unmatched="fill", fill_mapping=fill_mapping
        )
        self.assertEqual(result, {"name": "John", "age": 0, "is_student": False})

    def test_case_insensitive_matching(self):
        keys = ["Name", "Age", "Location"]
        dict_ = {"name": "John", "age": 30, "location": "NYC"}
        result = validate_keys(dict_, keys)
        self.assertEqual(result, {"Name": "John", "Age": 30, "Location": "NYC"})

    def test_special_characters(self):
        keys = ["name", "age", "email"]
        dict_ = {"name": "John", "age": 30, "email_address": "john@example.com"}
        result = validate_keys(dict_, keys)
        self.assertEqual(
            result, {"name": "John", "age": 30, "email": "john@example.com"}
        )

    def test_nested_dictionaries(self):
        keys = ["name", "age", "address"]
        dict_ = {
            "name": "John",
            "age": 30,
            "location": {"city": "New York", "country": "USA"},
        }
        result = validate_keys(dict_, keys)
        self.assertEqual(
            result,
            {
                "name": "John",
                "age": 30,
                "address": {"city": "New York", "country": "USA"},
            },
        )


class TestMdToJson(unittest.TestCase):
    def test_valid_markdown_json(self):
        md_content = """
        Here is a JSON example:
        ```json
        {"key": "value"}
        ```
        """
        self.assertEqual(md_to_json(md_content), {"key": "value"})

    def test_missing_keys(self):
        md_content = """
        ```json
        {"key": "value"}
        ```
        """
        with self.assertRaises(ValueError) as cm:
            md_to_json(md_content, expected_keys=["missing_key"])
        self.assertEqual(
            str(cm.exception), "Missing expected keys in JSON object: missing_key"
        )

    def test_custom_parser(self):
        def custom_parser(json_str):
            return {"parsed": True}

        md_content = """
        ```json
        {"key": "value"}
        ```
        """
        self.assertEqual(md_to_json(md_content, parser=custom_parser), {"parsed": True})

    def test_no_json_block(self):
        md_content = """
        Here is some text without a JSON block.
        """
        with self.assertRaises(ValueError) as cm:
            md_to_json(md_content)
        self.assertEqual(
            str(cm.exception), "No JSON code block found in the Markdown content."
        )


class TestEscapeCharsInJson(unittest.TestCase):
    def test_escape_default_chars(self):
        self.assertEqual(
            escape_chars_in_json('{"key": "value\n"}'), '{"key": "value\\n"}'
        )

    def test_custom_char_map(self):
        custom_map = {"\n": "\\n", "\r": "\\r"}
        self.assertEqual(
            escape_chars_in_json('{"key": "value\r\n"}', char_map=custom_map),
            '{"key": "value\\r\\n"}',
        )


class TestExtractJsonBlock(unittest.TestCase):
    def test_extract_json_block(self):
        md_content = """
        ```json
        {
            "key": "value",
            "key2": ["value1", "value2"],
            "key3": {"nested_key": "nested_value"}
        }
        ```
        """
        self.assertEqual(
            extract_json_block(md_content),
            {
                "key": "value",
                "key2": ["value1", "value2"],
                "key3": {"nested_key": "nested_value"},
            },
        )

    def test_no_json_block(self):
        md_content = """
        Here is some text without a JSON block.
        """
        with self.assertRaises(ValueError) as cm:
            extract_json_block(md_content)
        self.assertEqual(
            str(cm.exception), "No JSON code block found in the Markdown content."
        )

    def test_malformed_json_block(self):
        md_content = """
        ```json
        {"key": "value"
        ```
        """
        self.assertEqual(extract_json_block(md_content), {"key": "value"})

    def test_multiple_json_blocks(self):
        md_content = """
        ```json
        {"key1": "value1"}
        ```
        Some text in between.
        ```json
        {"key2": "value2"}
        ```
        """
        self.assertEqual(extract_json_block(md_content), {"key1": "value1"})

    def test_md_to_json(self):
        """Test extracting and validating JSON from Markdown content."""
        markdown = '```json\n{"key": "value"}\n```'
        result = md_to_json(markdown, expected_keys=["key"])
        self.assertEqual(result, {"key": "value"})
        with self.assertRaises(ValueError):
            md_to_json(markdown, expected_keys=["missing_key"])


class TestFuzzyParseJson(unittest.TestCase):
    def test_valid_json(self):
        self.assertEqual(fuzzy_parse_json('{"key": "value"}'), {"key": "value"})

    def test_single_quotes(self):
        self.assertEqual(fuzzy_parse_json("{'key': 'value'}"), {"key": "value"})

    def test_fix_brackets(self):
        self.assertEqual(fuzzy_parse_json('{"key": "value"'), {"key": "value"})

    def test_invalid_json(self):
        with self.assertRaises(ValueError):
            fuzzy_parse_json('{"key": "value",}')

    def test_empty_string(self):
        with self.assertRaises(ValueError):
            fuzzy_parse_json("")

    def test_incomplete_json(self):
        with self.assertRaises(ValueError):
            fuzzy_parse_json('{"key": ')

    def test_array_json(self):
        self.assertEqual(fuzzy_parse_json('["value1", "value2"]'), ["value1", "value2"])

    def test_nested_json(self):
        self.assertEqual(
            fuzzy_parse_json('{"key": {"nested_key": "nested_value"}}'),
            {"key": {"nested_key": "nested_value"}},
        )

    def test_extra_closing_bracket(self):
        with self.assertRaises(ValueError):
            fuzzy_parse_json('{"key": "value"}}')

    def test_missing_opening_bracket(self):
        with self.assertRaises(ValueError):
            fuzzy_parse_json('"key": "value"}')

    def test_mismatched_brackets(self):
        with self.assertRaises(ValueError):
            fuzzy_parse_json('{"key": [1, 2, 3}')

    def test_strict_mode(self):
        self.assertEqual(fuzzy_parse_json("{'key': 'value'}"), {"key": "value"})

    def test_basic_json_parsing(self):
        """Test parsing of correctly formatted JSON strings."""
        json_str = '{"name": "John", "age": 30}'
        expected = {"name": "John", "age": 30}
        self.assertEqual(fuzzy_parse_json(json_str), expected)

    def test_missing_closing_brackets(self):
        """Test JSON strings missing various closing brackets."""
        test_cases = [
            ('{"name": "John", "age": 30', {"name": "John", "age": 30}),
            ('[{"name": "John", "age": 30}', [{"name": "John", "age": 30}]),
            ('{"people": [{"name": "John"', {"people": [{"name": "John"}]}),
        ]
        for json_str, expected in test_cases:
            with self.subTest(json_str=json_str):
                self.assertEqual(fuzzy_parse_json(json_str), expected)

    def test_nested_structures(self):
        """Test with nested JSON objects and arrays."""
        json_str = '{"data": [{"id": 1, "value": "A"}, {"id": 2, "value": "B"'
        expected = {"data": [{"id": 1, "value": "A"}, {"id": 2, "value": "B"}]}
        self.assertEqual(fuzzy_parse_json(json_str), expected)

    def test_strict_mode(self):
        """Test the function's behavior in strict mode."""
        correct_json_str = '{"name": "John", "age": 30}'
        malformed_json_str = '{"name": "John", "age": 30'
        # Should work fine with correct JSON
        self.assertEqual(
            fuzzy_parse_json(correct_json_str),
            {"name": "John", "age": 30},
        )

    def test_error_handling(self):
        """Test with irreparably malformed JSON."""
        json_str = '{"name": "John", "age": 30 something wrong}'
        with self.assertRaises(ValueError):
            fuzzy_parse_json(json_str)


class TestForceValidateMapping(unittest.TestCase):
    def test_valid_json_string(self):
        keys = ["key1", "key2"]
        input_str = '{"key1": "value1", "key2": "value2"}'

        result = validate_mapping(input_str, keys)
        expected = {"key1": "value1", "key2": "value2"}
        self.assertEqual(result, expected)

    def test_single_quotes_json(self):
        keys = ["key1", "key2"]
        input_str = "{'key1': 'value1', 'key2': 'value2'}"

        result = validate_mapping(input_str, keys)
        expected = {"key1": "value1", "key2": "value2"}
        self.assertEqual(result, expected)

    def test_already_valid_dict(self):
        keys = ["key1", "key2"]
        input_dict = {"key1": "value1", "key2": "value2"}

        result = validate_mapping(input_dict, keys)
        expected = {"key1": "value1", "key2": "value2"}
        self.assertEqual(result, expected)

    def test_invalid_md_invalid_json(self):
        keys = ["key1", "key2"]
        input_str = """
        key1: value1,
        key2: value2,
        """
        with self.assertRaises(ValueError):
            validate_mapping(input_str, keys)

    def test_missing_keys(self):
        keys = ["key1", "key3"]
        input_str = '{"key1": "value1", "key2": "value2"}'

        with self.assertRaises(ValueError):
            validate_mapping(input_str, keys, strict=True)

    def test_extra_keys_in_input(self):
        keys = ["key1", "key2"]
        input_str = '{"key1": "value1", "key2": "value2", "key3": "value3"}'

        result = validate_mapping(input_str, keys, handle_unmatched="remove")
        expected = {"key1": "value1", "key2": "value2"}
        self.assertEqual(result, expected)

    def test_no_conversion_needed(self):
        keys = {"key1": "integer", "key2": "string"}
        input_dict = {"key1": 1, "key2": "value2"}

        result = validate_mapping(input_dict, keys)
        expected = {"key1": 1, "key2": "value2"}
        self.assertEqual(result, expected)

    def test_empty_input(self):
        keys = ["key1", "key2"]
        input_str = ""

        with self.assertRaises(ValueError):
            validate_mapping(input_str, keys)


if __name__ == "__main__":
    unittest.main()



from lion_core.communication.action_request import ActionRequest
from lion_core.libs import fuzzy_parse_json, nget, to_dict


def extract_request_plain_function_calling(
    response: dict,
    suppress=False,
) -> list[dict]:
    """Handles the action request parsing from the response."""
    try:
        tool_count = 0
        func_list = []
        while tool_count < len(response["tool_calls"]):
            _path = ["tool_calls", tool_count, "type"]

            if nget(response, _path) == "function":
                _path1 = ["tool_calls", tool_count, "function", "name"]
                _path2 = ["tool_calls", tool_count, "function", "arguments"]

                func_content = {
                    "action": f"action_{nget(response, _path1)}",
                    "arguments": nget(response, _path2),
                }
                func_list.append(func_content)
            tool_count += 1
        return func_list
    except:
        if not suppress:
            raise ValueError(
                "Response message must be one of regular response or function calling"
            )


def extract_request_from_content_code_block(content_: list[dict]) -> list[dict]:
    out = {}

    def _inner(request_):
        if "recipient_name" in request_:
            out["action"] = request_["recipient_name"].split(".")[1]
        elif "function" in request_:
            out["action"] = request_["function"]

        if "parameters" in request_:
            out["arguments"] = request_["parameters"]
        elif "arguments" in request_:
            out["arguments"] = request_["arguments"]

        if isinstance(_arg := out.get("arguments"), str):
            if (
                a := to_dict(_arg, str_type="json", parser=fuzzy_parse_json)
            ) is not None:
                out["arguments"] = a
            elif (a := to_dict(_arg, str_type="xml")) is not None:
                out["arguments"] = a

        _f = lambda x: x.replace("action_", "").replace("recipient_", "")
        return {"func": _f(out["action"]), "arguments": out["arguments"]}

    return [_inner(i) for i in content_]


def extract_action_request(content_: list[dict]) -> list[ActionRequest]:
    outs = []
    for request_ in content_:
        if "recipient_name" in request_:
            request_["action"] = request_.pop("recipient_name").split(".")[1]
        elif "function" in request_:
            request_["action"] = request_.pop("function")

        if "parameters" in request_:
            request_["arguments"] = request_["parameters"]
        elif "arguments" in request_:
            request_["arguments"] = request_["arguments"]

        f = lambda x: x.replace("action_", "").replace("recipient_", "")

        msg = ActionRequest(
            func=f(request_["action"]),
            arguments=request_["arguments"],
        )
        outs.append(msg)

    return outs



import asyncio
import unittest
from unittest.mock import AsyncMock, patch

from lion_core.libs.function_handlers._rcall import rcall


async def async_func(x: int) -> int:
    await asyncio.sleep(0.1)
    return x + 1


def sync_func(x: int) -> int:
    return x + 1


async def async_func_with_error(x: int) -> int:
    await asyncio.sleep(0.1)
    if x == 3:
        raise ValueError("mock error")
    return x + 1


def sync_func_with_error(x: int) -> int:
    if x == 3:
        raise ValueError("mock error")
    return x + 1


async def mock_handler(e: Exception) -> str:
    return f"handled: {str(e)}"


class TestRCallFunction(unittest.IsolatedAsyncioTestCase):
    async def test_rcall_basic(self):
        result = await rcall(async_func, 1, retries=2)
        self.assertEqual(result, 2)

    async def test_rcall_with_retries(self):
        result = await rcall(async_func_with_error, 3, retries=1, default=0)
        self.assertEqual(result, 0)

    async def test_rcall_with_timeout(self):
        with self.assertRaises(RuntimeError):
            await rcall(async_func, 1, timeout=0.05)

    async def test_rcall_with_error_handling(self):
        error_map = {ValueError: mock_handler}
        result = await rcall(async_func_with_error, 3, retries=1, error_map=error_map)
        self.assertEqual(result, "handled: mock error")

    async def test_rcall_with_timing(self):
        result, duration = await rcall(async_func, 1, timing=True)
        self.assertEqual(result, 2)
        self.assertTrue(duration > 0)

    async def test_rcall_with_initial_delay(self):
        with patch("asyncio.sleep", new_callable=AsyncMock) as mock_sleep:
            await rcall(async_func, 1, initial_delay=0.5)
            mock_sleep.assert_any_call(0.5)

    async def test_rcall_with_backoff_factor(self):
        with patch("asyncio.sleep", new_callable=AsyncMock) as mock_sleep:
            await rcall(
                async_func_with_error,
                3,
                retries=2,
                delay=0.1,
                backoff_factor=2,
                default=0,
            )
            mock_sleep.assert_any_call(0.1)
            mock_sleep.assert_any_call(0.2)

    async def test_rcall_with_kwargs(self):
        async def async_func_with_kwargs(x: int, add: int) -> int:
            await asyncio.sleep(0.1)
            return x + add

        result = await rcall(async_func_with_kwargs, 1, retries=2, add=2)
        self.assertEqual(result, 3)

    async def test_rcall_with_verbose(self):
        with patch("builtins.print") as mock_print:
            await rcall(async_func_with_error, 3, retries=1, verbose=True, default=0)
            mock_print.assert_any_call("Attempt 1/2 failed: mock error, retrying...")


if __name__ == "__main__":
    unittest.main()



from typing import Any

from pydantic import Field

from lion_core import event_log_manager
from lion_core.abc import Action
from lion_core.action.status import ActionStatus
from lion_core.exceptions import LionAccessError
from lion_core.generic.element import Element
from lion_core.generic.log import BaseLog


class ObservableAction(Element, Action):
    """
    Represents an action that can be observed, with a trackable status.

    The `ObservableAction` class extends `Action` and `Element` to provide a
    structure for actions whose status can be monitored. It includes a status
    field to indicate the current state of the action, and a request property
    that can be used to check for permission before the action is performed.

    Attributes:
        status (ActionStatus): The current status of the action. Defaults to `ActionStatus.PENDING`.

    Properties:
        request (dict): A dictionary representing the request for permission to perform the action, if any.
    """

    status: ActionStatus = ActionStatus.PENDING
    execution_time: float = None
    response: Any = None
    error: str = None
    retry_config: dict = Field(default_factory=dict, exclude=True)
    content_fields: list = ["response"]

    def __init__(self, retry_config: dict = None):
        super().__init__()
        self.retry_config = retry_config or {}

    @property
    def request(self) -> dict:
        """
        The request to get permission for, if any.

        This property returns a dictionary representing the request for
        permission before the action is performed. It can be overridden in
        subclasses to provide specific request details.

        Returns:
            dict: A dictionary containing the request details.
        """
        return {}

    async def alog(self) -> BaseLog:
        await event_log_manager.alog(self.to_log())

    def to_log(self) -> BaseLog:
        dict_ = self.to_dict()
        content = {k: dict_[k] for k in self.content_fields if k in dict_}
        loginfo = {k: dict_[k] for k in dict_ if k not in self.content_fields}
        return BaseLog(content=content, loginfo=loginfo)

    @classmethod
    def from_dict(cls, data):
        raise LionAccessError(
            "An action cannot be recreated. Once it's done, it's done."
        )


__all__ = ["ObservableAction"]
# File: lion_core/action/base.py



from lion_core.action.action_executor import ActionExecutor
from lion_core.action.action_processor import ActionProcessor
from lion_core.action.function_calling import FunctionCalling
from lion_core.action.tool import Tool, func_to_tool
from lion_core.action.tool_manager import ToolManager

__all__ = [
    "Tool",
    "func_to_tool",
    "FunctionCalling",
    "ToolManager",
    "ActionProcessor",
    "ActionExecutor",
]



import pytest

from lion_core.libs.data_handlers._nfilter import nfilter


@pytest.mark.parametrize(
    "data, condition, expected",
    [
        (
            {"a": 1, "b": {"c": 2, "d": 3}, "e": [4, 5, 6]},
            lambda x: isinstance(x, int) and x % 2 == 0,
            {"b": {"c": 2}, "e": [4, 6]},
        ),
        (
            [1, 2, [3, 4], {"a": 5, "b": 6}],
            lambda x: isinstance(x, int) and x % 2 == 0,
            [2, [4], {"b": 6}],
        ),
        (
            {"a": [1, {"b": 2, "c": [3, 4]}], "d": 5},
            lambda x: isinstance(x, int) and x % 2 == 0,
            {"a": [{"b": 2, "c": [4]}]},
        ),
        ({}, lambda x: True, {}),
        ([], lambda x: True, []),
        ({"a": 1, "b": 2, "c": 3}, lambda x: x > 1, {"b": 2, "c": 3}),
        ([1, 2, 3, 4, 5], lambda x: x % 2 == 0, [2, 4]),
        (
            {"a": 1, "b": "string", "c": [2, 3], "d": {"e": 4}},
            lambda x: isinstance(x, int),
            {"a": 1, "c": [2, 3], "d": {"e": 4}},
        ),
        (
            [1, "string", [2, 3], {"a": 4}],
            lambda x: isinstance(x, int),
            [1, [2, 3], {"a": 4}],
        ),
        (
            {"a": [1, 2], "b": {"c": 3, "d": 4}, "e": 5},
            lambda x: x % 2 != 0 if isinstance(x, int) else True,
            {"a": [1], "b": {"c": 3}, "e": 5},
        ),
        (
            [1, [2, [3, [4]]]],
            lambda x: x % 2 != 0 if isinstance(x, int) else True,
            [1, [[3, []]]],
        ),
        (
            {"a": 1, "b": 2, "c": {"d": 3, "e": 4, "f": {"g": 5, "h": 6}}},
            lambda x: x % 2 != 0 if isinstance(x, int) else True,
            {"a": 1, "c": {"d": 3, "f": {"g": 5}}},
        ),
        (
            [1, 2, 3, [4, 5, [6, 7, 8]]],
            lambda x: x > 3 if isinstance(x, int) else True,
            [[4, 5, [6, 7, 8]]],
        ),
        (
            {"a": "hello", "b": [1, "world", 3], "c": {"d": "python", "e": 2}},
            lambda x: isinstance(x, str),
            {"a": "hello", "b": ["world"], "c": {"d": "python"}},
        ),
        (
            [{"a": 1, "b": "x"}, {"c": 2, "d": "y"}, {"e": 3, "f": "z"}],
            lambda x: isinstance(x, str),
            [{"b": "x"}, {"d": "y"}, {"f": "z"}],
        ),
    ],
)
def test_nfilter(data, condition, expected):
    assert nfilter(data, condition) == expected


def test_nfilter_with_empty_result():
    data = {"a": 1, "b": 2, "c": 3}
    condition = lambda x: x > 10
    assert nfilter(data, condition) == {}


def test_nfilter_with_all_true_condition():
    data = {"a": 1, "b": {"c": 2, "d": 3}, "e": [4, 5]}
    condition = lambda x: True
    assert nfilter(data, condition) == data


def test_nfilter_with_nested_empty_structures():
    data = {"a": [], "b": {}, "c": {"d": []}}
    condition = lambda x: isinstance(x, (list, dict))
    assert nfilter(data, condition) == data


def test_nfilter_with_custom_objects():
    class CustomObj:
        def __init__(self, value):
            self.value = value

    data = {"a": CustomObj(1), "b": CustomObj(2), "c": [CustomObj(3), CustomObj(4)]}
    condition = lambda x: isinstance(x, CustomObj) and x.value % 2 == 0
    expected = {"b": CustomObj(2), "c": [CustomObj(4)]}
    result = nfilter(data, condition)
    assert isinstance(result["b"], CustomObj) and result["b"].value == 2
    assert isinstance(result["c"][0], CustomObj) and result["c"][0].value == 4


def test_nfilter_with_none_values():
    data = {"a": None, "b": 1, "c": {"d": None, "e": 2}}
    condition = lambda x: x is not None
    expected = {"b": 1, "c": {"e": 2}}
    assert nfilter(data, condition) == expected


def test_invalid_input():
    with pytest.raises(
        TypeError, match="The nested_structure must be either a dict or a list."
    ):
        nfilter(42, lambda x: True)


def test_nfilter_with_exception_in_condition():
    def faulty_condition(x):
        if isinstance(x, int):
            raise ValueError("Error in condition")
        return True

    data = {"a": 1, "b": "string", "c": [2, 3]}
    with pytest.raises(ValueError, match="Error in condition"):
        nfilter(data, faulty_condition)


# File: tests/test_nfilter.py
